%% Main function for the Category Covering process
function [ok,raw_data,processed_data,results] = vns_function(filename,n,minK,maxK,runs,maxIterations,timeLimit,display_interval,random_seed,handles)
%% Inputs:
% filename: input file name (.xlsx or .csv)
% n: the program will count only piles that have been made at least 'n' times
% minK: minimum value for the number of summary piles
% maxK: maximum value for the number of summary piles
% runs: number of replications
% maxIterations: maximum VNS iterations per execution
% timeLimit: time limit for each execution
% display_interval: solutions will be displayed between 'display_interval' iterations
% random_seed: base seed for random number genetarion
% handles: handles for GUI objects

%% Outputs:
% ok: true if there's no error when reading file, false otherwise
% raw_data: struct with raw input data
% processed_data: struct with weighted piles without repetitions
% results: struct with summary piles (pmj), coverings (ximli), solution cost (Z), statistics and others

%% Checking options and reading file
ok = true;
% Checking if piles from the same individual must be covered by different
% summay piles. If that's true, then functions terminated with 'OPPSP' are
% called.
if get(handles.cBonePilePerSumPile,'Value') == 0
    onePilePerSumPile = false;
else
    onePilePerSumPile = true;
end
% Reading file
set_message(handles,'Reading file...');
if get(handles.rBxlsx,'Value') == 1 % read .xlsx file
    [error,I,J,ci,yilij,weights,raw_data]=loadData_xlsx(filename,n);
else                                % read .csv file
    [error,I,J,ci,yilij,weights,raw_data]=loadData_csv(filename,n);
end
% Returning an error if present
if ~strcmp(error,'')
    set_message(handles,error,'error');
    ok = false;processed_data=-1;results=-1;
    return;
end
% Setting some results
processed_data.yilij = yilij;
processed_data.weights = weights;
% If onePilePerSumPile==true then the minimum number of summary piles is
% the highest number of piles created by individuals
if onePilePerSumPile
    yilij = raw_data.yilij;
    I = length(yilij);
    ci = zeros(I,1);
    for i = 1:I
        ci(i) = size(yilij{i},1);
    end
    [val,id] = max(ci);
    if val > minK
        set_message(handles,['The minimum number of summary piles for this sorting ',...
            'task data is ' num2str(val) ' because one consumer (ID=' num2str(raw_data.IDs(id)) ') has sorted this amount of piles.'],'error');
        ok = false;processed_data=-1;results=-1;
        return;
    end
end
% Checking if the log text can be scrolled down
can_scroll_down = true;
try
    jhEdit = findjobj(handles.eTlog);
    jEdit = jhEdit.getComponent(0).getComponent(0);
catch
    can_scroll_down = false;
    disp('Cannot scroll down');
end

%% Starting the algorithm
set_message(handles,'Starting...');
% Setting some results and configurations and initializing cells
for K = minK:maxK
    for run = 1:runs
        results.seed(K-minK+1,run) = random_seed + (run-1) + runs*(K-minK);
    end
end
results.minK = minK;
results.maxK = maxK;
results.runs=runs;
results.pmj = cell(maxK-minK+1,runs);
results.ximli = cell(maxK-minK+1,runs);
results.Z = zeros(maxK-minK+1,runs);
results.nIterations = zeros(maxK-minK+1,runs);
results.truePos = cell(maxK-minK+1,runs);
results.trueNeg = cell(maxK-minK+1,runs);
results.covRate = cell(maxK-minK+1,runs);
log_msg = '';
k_max = 20;
lastTimeLeft = inf;
% Starting algorithm for each number of summary piles and each replication
for K = minK:maxK
    for run = 1:runs
        rng(results.seed(K-minK+1,run)); % Random seed for this run
        log_msg = sprintf('%s\n--- K: %d - Run: %d ---\n\n',log_msg,K,run);
        set(handles.eTlog,'String',log_msg);
        drawnow
        % scrolls the log down if possible
        if can_scroll_down
            jEdit.select(length(log_msg)-1,length(log_msg)); % by selecting the last character it sets the caret position to the end
        end

        % VNS settings
        stop_condition=false;
        k=1;
        nIterations=1;
        
        % Initial solution: summary piles (pmj) and linking variables (ximli) are set to random
        pmj={randi(2,[K,J])-1};
        bestPmj=pmj;
        ximli=cell(I,1);
        for i=1:I
            ximli{i}=zeros(K,ci(i));
            if onePilePerSumPile
                perm = randperm(K,ci(i));
                for li=1:ci(i)
                    ximli{i}(perm(li),li)=1;
                end
            else
                for li=1:ci(i)
                    ximli{i}(randi(K),li)=1;
                end
            end
        end
        bestXimli=ximli;
        bestZ=inf;
        
        % Iterations or estimated time left
        runsLeft = runs-run+1+(maxK-K)*runs;
        if timeLimit ~= inf
            timeLeftForThisRun = ceil(runsLeft*timeLimit);
        elseif maxIterations ~= inf
            iterationsLeft = runsLeft*maxIterations;
        else
            set_message(handles,'','clear');
        end
        
        % Starting VNS iterations
        start=tic;
        while(stop_condition==false)
            % Iterations or estimated time left
            if timeLimit ~= inf
                timeLeft = ceil(max(timeLeftForThisRun - toc(start),0));
                if timeLeft ~= lastTimeLeft
                    lastTimeLeft = timeLeft;
                    set_message(handles,['Estimated time left: ' datestr(timeLeft/86400, 'HH:MM:SS')]);
                end
            elseif maxIterations ~= inf
                if mod(iterationsLeft-nIterations,50)==0
                    set_message(handles,['Iterations left: ', num2str(iterationsLeft-nIterations)]);
                end
            end
            
            % Shaking and local search
            if onePilePerSumPile
                ximli=shakingOPPSP(bestXimli,bestPmj,yilij,k);
                [Z,pmj,ximli]=local_searchOPPSP(ximli,pmj,yilij);
            else
                ximli=shaking(bestXimli,bestPmj,yilij,k);
                [Z,pmj,ximli]=local_search(ximli,pmj,yilij,weights);
            end
            % Updating best solution
            if Z < bestZ
                bestZ=Z;
                bestPmj=pmj;
                bestXimli=ximli;
                time = toc(start);
                log_msg = sprintf('%sIteration: %d - Best solution: %d - %f seconds elapsed.\n',log_msg,nIterations,bestZ,time);
                set(handles.eTlog,'String',log_msg);
                drawnow
                if can_scroll_down
                    jEdit.select(length(log_msg)-1,length(log_msg));
                end
                k=1;
            else
                k=mod(k,k_max)+1;
            end
            % Display if 'nIterations' is a multiple of 'display_interval'
            if mod(nIterations,display_interval)==0
                time = toc(start);
                log_msg = sprintf('%sIteration: %d - Best solution: %d - %f seconds elapsed.\n',log_msg,nIterations,bestZ,time);
                set(handles.eTlog,'String',log_msg);
                drawnow
                if can_scroll_down
                    jEdit.select(length(log_msg)-1,length(log_msg));
                end
            end
            % Check if stop conditions are met
            if toc(start)>timeLimit || nIterations == maxIterations
                stop_condition=true;
                time = toc(start);
                log_msg = sprintf('%sIteration: %d - Best solution: %d - %f seconds elapsed.\n',log_msg,nIterations,bestZ,time);
                set(handles.eTlog,'String',log_msg);
                drawnow
                if can_scroll_down
                    jEdit.select(length(log_msg)-1,length(log_msg));
                end
            end
            nIterations = nIterations+1;
        end
        
        % Check if the best solution cost is trully the cost of this solution
        if onePilePerSumPile
            [~,Z]=get_costOPPSP(bestPmj,yilij);
        else
            [~,Z,~]=get_cost(bestPmj,yilij,weights);
        end
        if Z~=bestZ
            disp('Error!')
        end
        
        % Check if summary piles are equal to any pile
        if onePilePerSumPile
            text = check_pilesOPPSP(bestPmj,yilij,bestXimli);
        else
            text = check_piles(bestPmj,yilij,bestXimli,weights,raw_data.yilij);
        end
        log_msg = sprintf('%s%s',log_msg,text);
        set(handles.eTlog,'String',log_msg);
        drawnow
        jEdit.select(length(log_msg)-1,length(log_msg));
        
        % Compute some statistics and set results
        if onePilePerSumPile
            [truePos,trueNeg,covRate] = ratingOPPSP(bestPmj,yilij,bestXimli);
        else
            [truePos,trueNeg,covRate] = rating(bestPmj,yilij,bestXimli,weights);
        end
        results.pmj{K-minK+1,run} = bestPmj{1};
        if onePilePerSumPile
            results.ximli{K-minK+1,run} = bestXimli;
        else
            results.ximli{K-minK+1,run} = bestXimli{1};
        end
        results.Z(K-minK+1,run) = bestZ;
        results.nIterations(K-minK+1,run) = nIterations-1;
        results.truePos{K-minK+1,run} = truePos;
        results.trueNeg{K-minK+1,run} = trueNeg;
        results.covRate{K-minK+1,run} = covRate;
    end 
end
end